#![allow(unused_imports)]

use euclid::Point2D;
use ndarray::prelude::*;

mod point;
mod math_characteristic_functions;
mod math_levelset_functions;
use crate::point::*;
use crate::math_characteristic_functions::*;
use crate::math_levelset_functions::*;

//Structure of File:
//nbhd_check -> Defines the rule for the local neighborhood (kernel)
//  dist:f64,r:f64 -> bool

//create_normals -> Computes the normals for a uniform sample of points
//  m: usize -> Vec<(f64, f64)>

pub fn nbhd_check(dist:f64,r:f64)->bool{
    //Ball
    dist<=r
    
    //Annulus
    //dist<=r && dist>=r/2.0
}

fn create_normals(
    m: usize,
) -> (Vec<(f64, f64)>, f64)
{
    let angles = Array::linspace(0., 2.0 * std::f64::consts::PI, m + 1);
    let dphi = 2.0 * std::f64::consts::PI / (m as f64);
    let mut normal: Vec<(f64, f64)> = Vec::new();
    for i in 0..m {
        normal.push((angles[i].cos(), angles[i].sin()));
    }
    (normal, dphi)
}

fn main() {
    //mnist_test::mnist_main();

    //let mut v = [-5f64, 4f64, 2f64, -3f64, 1f64];
    //-5 -3 1 2 4

    //let (_lesser, median, _greater) = v.select_nth_unstable_by(3, |a, b| b.total_cmp(a));
    //println!("select: {}", *median as f64);

    //0 -> 4
    //1 -> 2
    //2 -> 1
    //3 -> -3
    //4 -> -5
    //5 -> PANIK
    //println!("select: {}", median);

    //anteil=cos(alpha)*|Br Dc|/|Br|
    //fuer x<r oder x>1-r nun x delta ist flaeche arccos(x/r)-sqrt(r^2-x^2)x (hinterer teil ist r^2 mal sin(2alpha)/2=sin cos)
    //beiden beiden ist es additiv falls a_x+a_y<pi/2 
    //sonst beta=(a_x+a_y-pi/2), siehe unten
    
    // let x=0.1;
    // let y=0.05;
    // let r=0.2;
    // let dx=min(x, 1-x), dy=min(y, 1-x);
    //let ax=0; 
    //let ay=0;
    //let volx=0;
    //let voly=0;
    // if(dx<r)ax=arccos(dx/r); volx=ax*r*r-sqrt(r*r-dx*dx)*dx
    // if(dy<r)ay=arccos(dy/r); voly=ay*r*r-sqrt(r*r-dy*dy)*dy
    // if(dx<r & dy<r & ax+ay>pi/2) (volx+voly-pi/2*r*r)/2+dx*dy
    //println!("{}",PI);
    //println!("{}",0.84f64.acos());
    //println!("{}", (std::f64::consts::PI/4.).cos());

    //cos of angle
    //let angle : f64 = (std::f64::consts::PI*0./180.).cos();//-0.71;
    //Ob in i32 oder f64 mit chunks/ohne gearbeitet wird
    let chunks_bool : bool=false;
    let loadout: (usize, f64, i32, usize, f64, usize) = (
        200_000_usize,//200
        0.05_f64,
        1500_i32,
        30_usize,//20
        0.04_f64,
        100_usize,
    );

    let n: usize = std::env::args().nth(1).expect("no number of points given").parse::<usize>().unwrap();
    let r: f64 = std::env::args().nth(2).expect("no radius of kernel given").parse::<f64>().unwrap();
    let it: i32 = std::env::args().nth(3).expect("no number of iterations given").parse::<i32>().unwrap();
    //let m: usize = std::env::args().nth(5).expect("no number of discretization points given").parse::<usize>().unwrap();
    //let r_min: f64 = std::env::args().nth(6).expect("no minimal distance given").parse::<f64>().unwrap();
    let chunk_number: usize = std::env::args().nth(4).expect("no number of chunks given").parse::<usize>().unwrap();
    let angle: f64 = (std::f64::consts::PI*(std::env::args().nth(5).unwrap_or("90.0".to_string()).parse::<f64>().unwrap())/180.0).cos();
    let m: usize = std::env::args().nth(6).unwrap_or("10".to_string()).parse::<usize>().unwrap();
    let euler: usize = std::env::args().nth(7).unwrap_or("10".to_string()).parse::<usize>().unwrap();
    let r_min: f64 = std::env::args().nth(8).unwrap_or("0.04".to_string()).parse::<f64>().unwrap();

    println!("Loadout (N, r, it, n_chunk, m, r_min) = ({:?}, {:?}, {:?}, {:?}, {:?}, {:?})", n, r, it, chunk_number, m, r_min);

    let (normal, dphi): (Vec<(f64, f64)>, f64)=create_normals(m);
    let mut ellipse: Vec<(f64, f64)>= vec![(1.,1.); m];
    let mut p: Vec<f64> = vec![1.; m];
    let mut dp: Vec<f64> = vec![1.; m];
    let mut ddp: Vec<f64> = vec![1.; m];
    //let mut dp: [f64; M] = [1.0; M];
    // let mut ddp: [f64; M] = [1.0; M];
    let asqr = 8./75.;//4. / 50.0;
    let bsqr = 2./75.;//1. / 50.0;
    // const MP: usize = M - 1;
    let mp=m-1;
    let dt: f64 = r * r*1. / 6.;// 1/6 kreis,7/24 annulus
    //heat 1/8?

    if !chunks_bool {//f64
        let mut points = generate_random_pointsf64(n);
        println!("Values generated");
    
        //Reference domain picture for nbhd drawn
        let base=Point2D::new(0.1, 0.1);
        let mut reference_points=Vec::new();
        for i in 0..points.len() {
            if nbhd_check(points[i].position.distance_to(base), r)
                {reference_points.push(i);}
        }

        //Computes the neighbors for all points and saves them in a Hashmap point->nbhd(point)
        let neighbors = 
            compute_neighborsf64(&points, r);
        //Computes the p of p-median for Young angles
        let p_values = compute_p_valuesf64(&points, r, angle);
        println!("Neighborhoods computed");
    
        //Creates structure of the images
        let folder = "imgs";
        std::fs::create_dir_all(folder).unwrap();
    
        //Ellipse plotten
        for j in 0..m {
            // Tangentiale Polar
            p[j] = (asqr * normal[j].0.powi(2) + bsqr * normal[j].1.powi(2)).sqrt();//0.4
            //p[j] = 0.4-0.1*normal[j].1.powi(8);//0.4
            ellipse[j] = (
                0.5 + p[j] * normal[j].0 - dp[j] * normal[j].1,
                0.5 + p[j] * normal[j].1 + dp[j] * normal[j].0,
            );//

            /*Polarcoords
            p[j] = 0.4-0.32*normal[j].1.powi(2);
            ellipse[j] = (
                0.5 + p[j] * normal[j].0,
                0.5 + p[j] * normal[j].1,
            );*/
        }
        for j in 0..m {
            match j {
                0 => {
                    dp[0] = (p[1] - p[mp]) / (2. * dphi);
                    ddp[0] = (p[1] + p[mp] - 2. * p[0]) / (dphi * dphi);
                }
                x if x==mp => {
                    dp[mp] = (p[0] - p[mp-1]) / (2. * dphi);
                    ddp[mp] = (p[0] + p[mp-1] - 2. * p[mp]) / (dphi * dphi);
                }
                _ => {
                    dp[j] = (p[j + 1] - p[j - 1]) / (2. * dphi);
                    ddp[j] = (p[j + 1] + p[j - 1] - 2. * p[j]) / (dphi * dphi);
                }
            }

            // Tangentiale Polar
            ellipse[j] = (
                0.5 + p[j] * normal[j].0 - dp[j] * normal[j].1,
                0.5 + p[j] * normal[j].1 + dp[j] * normal[j].0,
            );//
            /*Polarcoords
            ellipse[j] = (
                0.5 + p[j] * normal[j].0,
                0.5 + p[j] * normal[j].1,
            );*/
        }
    
        //Render Image
        render_to_imagef64(
            &points,
            &(format!("{}/{:03}.png", folder, 0)),
            &reference_points,
            0.,
            &ellipse,
        );
    
        println!("First Frame complete");
    
        for i in 0..it {
            if i % 10 == 0 {
                println!("Iteration {}", i);
            }

            update_valuesf64(
                &mut points,
                &neighbors,
                &p_values,
            );

            // for j in 0..points.len() {
            //     if points[j].position.x<2.*r || points[j].position.y<2.*r || points[j].position.x>1.-2.*r || points[j].position.y>1.-2.*r
            //     {
            //         points[j].value=point_valuef64(points[j].position.x,points[j].position.y)+4.*((i+1) as f64*DT);
            //     }
            // }

            if p[0] > loadout.4 {
                for k in 0..euler{
                    for j in 0..m {

                        // Tangentiale Polar
                        //Zeit-Euler-Schema
                        p[j] = p[j] - dt / (euler as f64*(p[j] + ddp[j]));
                        //p[j] = (p[j].powi(2) - dt *2.).sqrt();
                        ellipse[j] = (
                            0.5 + p[j] * normal[j].0 - dp[j] * normal[j].1,
                            0.5 + p[j] * normal[j].1 + dp[j] * normal[j].0,
                        );//
                        /*Polarcoods
                        let normalx=p[j]*normal[j].0+dp[j]*normal[j].1;
                        let normaly=p[j]*normal[j].1-dp[j]*normal[j].0;
                        let norm=(normalx.powi(2)+normaly.powi(2)).sqrt();
                        let curv=((-2.*dp[j]*normal[j].1+normal[j].0*(ddp[j]-p[j])).powi(2)+(2.*dp[j]*normal[j].0+normal[j].1*(ddp[j]-p[j])).powi(2)).sqrt()/(norm.powi(3));
                        p[j] = p[j] - dt *;
                        ellipse[j] = (
                            0.5 + p[j] * normal[j].0,
                            0.5 + p[j] * normal[j].1,
                        );*/
                    }
                    for j in 0..m {
                        match j {
                            0 => {
                                dp[0] = (p[1] - p[mp]) / (2. * dphi);
                                ddp[0] = (p[1] + p[mp] - 2. * p[0]) / (dphi * dphi);
                            }
                            x if x==mp => {
                                dp[mp] = (p[0] - p[mp-1]) / (2. * dphi);
                                ddp[mp] = (p[0] + p[mp-1] - 2. * p[mp]) / (dphi * dphi);
                            }
                            _ => {
                                dp[j] = (p[j + 1] - p[j - 1]) / (2. * dphi);
                                ddp[j] = (p[j + 1] + p[j - 1] - 2. * p[j]) / (dphi * dphi);
                            }
                        }
                        ellipse[j] = (
                            0.5 + p[j] * normal[j].0 - dp[j] * normal[j].1,
                            0.5 + p[j] * normal[j].1 + dp[j] * normal[j].0,
                        );
                    } 
                }
            }
            if i % 1 == 0 {
                let filename = format!("{}/{:03}.png", folder, (i/1+1));
                render_to_imagef64(&points, &filename, &reference_points,(i+1) as f64*dt, &ellipse);
            }
        }
    
        println!("Finished");
    
        let output_video = "output.mp4";
        let ffmpeg_command = format!(
            "ffmpeg -framerate 2 -i {}/%03d.png -c:v libx264 -r 30 {}",
            folder, output_video
        );
        std::process::Command::new("sh")
            .arg("-c")
            .arg(&ffmpeg_command)
            .output()
            .expect("Failed to execute ffmpeg command");
    
        println!("Video saved as: {}", output_video);
    }else{
        let mut points = generate_random_points(n);
        println!("Values");
    
        //Reference for nbhd
        let base=Point2D::new(0.1, 0.1);
        let mut reference_points=Vec::new();
        for i in 0..points.len() {
            if nbhd_check(points[i].position.distance_to(base), r)
                {reference_points.push(i);}
        }
        let (chunk_map, neighbors, chunk_count)=
            compute_neighbors(&points, r, chunk_number);
    
        println!("Nbhds");
    
        let folder = "imgs";
        std::fs::create_dir_all(folder).unwrap();
    
        for j in 0..m {
            p[j] = 0.4;//(asqr * normal[j].0.powi(2) + bsqr * normal[j].1.powi(2)).sqrt();
            ellipse[j] =(
                0.5 + p[j] * normal[j].0,
                0.5 + p[j] * normal[j].1,
            );
        }
        /*for j in 0..M {
            match j {
                0 => {
                    dp[0] = (p[1] - p[M - 1]) / (2. * dphi);
                    ddp[0] = (p[1] + p[M - 1] - 2. * p[0]) / (dphi * dphi)s
                }
                MP => {
                    dp[M - 1] = (p[0] - p[M - 2]) / (2. * dphi);
                    ddp[M - 1] = (p[0] + p[M - 2] - 2. * p[M - 1]) / (dphi * dphi)
                }
                _ => {
                    dp[j] = (p[j + 1] - p[j - 1]) / (2. * dphi);
                    ddp[j] = (p[j + 1] + p[j - 1] - 2. * p[j]) / (dphi * dphi)
                }
            }
            ellipse[j] = (
                0.5 + p[j] * normal[j].0 - dp[j] * normal[j].1,
                0.5 + p[j] * normal[j].1 + dp[j] * normal[j].0,
            );
        }*/
        println!("PreConv");
    
        let mut conv_hull=convex_hull(
            &(points
                .iter()
                .filter_map(|&p| if p.value==0 {Some(p)} else {None}).collect())
        );
        println!("PostConv");
    
        render_to_image(
            &points,
            &(format!("{}/{:03}.png", folder, 0)),
            &ellipse,
            &conv_hull,
            &reference_points,
        );
    
        println!("First Frame");
    
        for i in 0..it {
            if i % 1 == 0 {
                println!("Iteration {}", i);
            }
            update_values(
                &mut points,
                &chunk_map,
                &neighbors,
                &chunk_count,
                chunk_number,
            );
            conv_hull=convex_hull(
                &(
                points
                .iter()
                .filter_map(|&p| {
                    if p.value==0
                    {Some(p)} else {None}
                }).collect())
            );
    
            if p[0] > loadout.4 {
                for j in 0..m {
                    //Zeit-Euler-Schema
                    //p[j] = p[j] - DT / (p[j] + ddp[j]);
                    p[j] = (p[j].powi(2) - dt *2.).sqrt();
                    ellipse[j] =(
                        0.5 + p[j] * normal[j].0,
                        0.5 + p[j] * normal[j].1,
                    );
                }
                /*for j in 0..M {
                    match j {
                        0 => {
                            dp[0] = (p[1] - p[M - 1]) / (2. * dphi);
                            ddp[0] = (p[1] + p[M - 1] - 2. * p[0]) / (dphi * dphi)
                        }
                        MP => {
                            dp[M - 1] = (p[0] - p[M - 2]) / (2. * dphi);
                            ddp[M - 1] = (p[0] + p[M - 2] - 2. * p[M - 1]) / (dphi * dphi)
                        }
                        _ => {
                            dp[j] = (p[j + 1] - p[j - 1]) / (2. * dphi);
                            ddp[j] = (p[j + 1] + p[j - 1] - 2. * p[j]) / (dphi * dphi)
                        }
                    }
                    ellipse[j] = (
                        0.5 + p[j] * normal[j].0 - dp[j] * normal[j].1,
                        0.5 + p[j] * normal[j].1 + dp[j] * normal[j].0,
                    );
                }*/
            }
            println!("distance {}: {:?}", i, distance_convert(&conv_hull, &ellipse).sqrt());
            if i % 1 == 0 {
                let filename = format!("{}/{:03}.png", folder, i/1+1);
                render_to_image(&points, &filename, &ellipse,&conv_hull, &reference_points);
            }
        }
    
        println!("Finished");
    
        let output_video = "output.mp4";
        let ffmpeg_command = format!(
            "ffmpeg -framerate 2 -i {}/%03d.png -c:v libx264 -r 30 {}",
            folder, output_video
        );
        std::process::Command::new("sh")
            .arg("-c")
            .arg(&ffmpeg_command)
            .output()
            .expect("Failed to execute ffmpeg command");
    
        println!("Video saved as: {}", output_video);
    }
}


//Werte
    // const loadout: (usize, f64, i32, usize, f64, usize) = (
    //     1_000_000_usize,//200
    //     0.01_f64,
    //     8000_i32,
    //     20_usize,//20
    //     0.04_f64,
    //     100_usize,
    // );
    //Pointnumber, r, it, ellipse discretiation, min_r, chunk_number
    // const loadout: (usize, f64, i32, usize, f64, usize) = (
    //     100_000_usize,//200
    //     0.1_f64,
    //     15_i32,
    //     10_usize,
    //     0.04_f64,
    //     20_usize,
    // );
    // const loadout: (usize, f64, i32, usize, f64, usize) = (
    //     150_000_usize,//200
    //     0.07_f64,
    //     50_i32,
    //     20_usize,//20
    //     0.04_f64,
    //     100_usize,
    // );
    // const loadout: (usize, f64, i32, usize, f64, usize) = (
    //     1_000_000_usize,
    //     0.03_f64,
    //     150_i32,
    //     30_usize,
    //     0.04_f64,
    //     100_usize,
    // );
    // const loadout: (usize, f64, i32, usize, f64, usize) = (
    //     10_000_000_usize,
    //     0.02_f64,
    //     320_i32,
    //     30_usize,
    //     0.01_f64,
    //     250_usize,
    // );

//Laedt eine beliebige Abfolge von Eingaben
    /*let args: Vec<String> = env::args().collect();

    match args.len() {
        // no arguments passed
        1 => {
            println!("My name is 'match_args'. Try passing some arguments!");
        },
        // one argument passed
        2 => {
            match args[1].parse() {
                Ok(42) => println!("This is the answer!"),
                _ => println!("This is not the answer."),
            }
        },
        // one command and one argument passed
        3 => {
            let cmd = &args[1];
            let num = &args[2];
            // parse the number
            let number: i32 = match num.parse() {
                Ok(n) => {
                    n
                },
                Err(_) => {
                    eprintln!("error: second argument not an integer");
                    help();
                    return;
                },
            };
            // parse the command
            match &cmd[..] {
                "increase" => increase(number),
                "decrease" => decrease(number),
                _ => {
                    eprintln!("error: invalid command");
                    help();
                },
            }
        },
        // all the other cases
        _ => {
            // show a help message
            help();
        }
    }
    */
