// mod point;
use euclid::Point2D;
use rayon::prelude::*;
use std::collections::HashMap;
use ndarray::prelude::*;
use image::{Rgb, RgbImage};
use imageproc::drawing::draw_line_segment_mut;
use crate::point::*;
use crate::nbhd_check;

//Structure of File:
//convex_hull -> Calculates the Convex Hull of given points
//  points: &Vec<DataPoint> -> Vec<DataPoint>

//distance_convert -> A streamlined version that combines distance and convert
//  first: &Vec<DataPoint>, second: &[(f64, f64)] -> f64

//distance -> Gives one-sided distance (first from second?) between two lines
//  (first,second): (Vec<DataPoint>,Vec<DataPoint>) -> f64

//line_dist -> Calculates the distance between one point and the line spanned by two others [used in distance]
//  first: Point2D<f64,EmptyUnit>, second1: Point2D<f64,EmptyUnit>, second2: Point2D<f64,EmptyUnit> -> f64

//convert -> Data type conversion
//  first: &Vec<DataPoint>, second: &[(f64, f64)] -> (Vec<DataPoint>,Vec<DataPoint>)

//compute_neighbors -> Computes nbhds and ???
//  points: &[DataPoint], radius: f64, chunknmb: usize -> HashMap<usize, Vec<(usize, usize)>>, HashMap<usize, Vec<usize>>, HashMap<(usize, usize), i32>

//update_values -> Updates the values in each step
//  points: &mut Vec<DataPoint>, chunk_map: &HashMap<usize, Vec<(usize, usize)>>, 
//  neighbors: &HashMap<usize, Vec<usize>>, chunk_count: &HashMap<(usize, usize), i32>, chunknmb: usize

//render_to_image -> Renders the image
//  points: &[DataPoint], filename: &str, ellipse: &[(f64, f64)], conv_hull : &Vec<DataPoint>, ref_points :&Vec<usize>


// The convex hull of a set of points is the smallest convex polygon that contains all the points.
pub fn convex_hull(
    points: &Vec<DataPoint>,
) -> Vec<DataPoint>
{
    Vec::new()
    /*if points.len()>0
    {
        let mut conv_hull : Vec<DataPoint> = Vec::new();
        // point with smallest x-coordinate as start value
        let first_point = points.into_iter().min_by(|a, b| a.position.x.partial_cmp(&b.position.x).unwrap()).unwrap();
        conv_hull.push(*first_point);

        //let current_point = first_point;
        loop {
            let current_point=conv_hull[conv_hull.len()-1];
            let cross_product = |p: &DataPoint, q: &DataPoint| -> f64 {
                (p.position.x - current_point.position.x) * (q.position.y - current_point.position.y) - (p.position.y - current_point.position.y) * (q.position.x - current_point.position.x)
            };
            let next_point = points.iter().
                min_by(|a, b| {
                    cross_product(*a, *b).partial_cmp(&0.0).unwrap()
                }).unwrap();
            if next_point.position == first_point.position {
                break;
            }
            conv_hull.push(*next_point);
        }

        // let mut it=0;//kann it==0 durch len constraint ersetzen
        // let mut temp_point=*first_point;

        // while it<10000 && (it==0 || !(temp_point.position.x==conv_hull[0].position.x && temp_point.position.y==conv_hull[0].position.y))
        // {
        //     it+=1;
        //     temp_point=points[0];
        //     let base_point=conv_hull[conv_hull.len()-1];

        //     let cross_product = |p: &DataPoint, q: &DataPoint| -> f64 {
        //         (p.position.x - base_point.position.x) * (q.position.y - base_point.position.y) - (p.position.y - base_point.position.y) * (q.position.x - base_point.position.x)
        //     };
        //     for j in 0..points.len() {
        //         if cross_product(&temp_point, &points[j])>0.0
        //             {temp_point=points[j];}
        //     }
        //     conv_hull.push(temp_point);
        // }
        conv_hull
    }else{
        Vec::new()
    }*/
}


pub fn distance_convert(
    first: &Vec<DataPoint>,
    second: &[(f64, f64)],
) -> f64
{
    distance(convert(first, second))
}

fn distance(
    (first,second): (Vec<DataPoint>,Vec<DataPoint>)
) -> f64
{
    let mut dist=-1.0;
    for j in 0..first.len() {//b=first-second1, a=second2-second1, b^2-(a*b)^2/a^2
        let mut inner_dist=line_dist(first[j].position, second[second.len()-1].position, second[0].position);
        for k in 0..second.len()-1 {
            inner_dist=inner_dist.min(line_dist(first[j].position, second[k].position, second[k+1].position));
        } 
        if inner_dist>dist {
            dist=inner_dist;
        }
    }
    dist
}

fn line_dist(first: Point2D<f64,EmptyUnit>, second1: Point2D<f64,EmptyUnit>, second2: Point2D<f64,EmptyUnit>)
    -> f64
{
    let vectorb=first-second1;
    let vectora=second2-second1;
    let prod=vectora.dot(vectorb);
    if prod<0.0f64
    {
        vectorb.square_length()
    } else {
        if prod>vectora.square_length()
        {
            (first-second2).square_length()
        }else {
            vectorb.square_length()-prod.powi(2)/vectora.square_length()
        }
    }
}

fn convert(
    first: &Vec<DataPoint>,
    second: &[(f64, f64)],
) -> (Vec<DataPoint>,Vec<DataPoint>)
{
    let mut second_conv : Vec<DataPoint>=Vec::new();
    for i in 0..second.len() {
        let (x,y)=second[i];
        second_conv.push(DataPoint { position: Point2D::new(x, y), value: point_value(x, y) });
    }
    (first.to_vec(), second_conv)
}

pub fn compute_neighbors(
    points: &[DataPoint],
    radius: f64,
    chunknmb: usize,
) -> (
    HashMap<usize, Vec<(usize, usize)>>,
    HashMap<usize, Vec<usize>>,
    HashMap<(usize, usize), i32>,
) {
    //chunks and nbhdpunkte die bleiben
    let mut neighbors_map: HashMap<usize, Vec<usize>> = HashMap::new();
    let mut chunk_map: HashMap<usize, Vec<(usize, usize)>> = HashMap::new();
    let mut chunk_count: HashMap<(usize, usize), i32> = HashMap::new();

    //Berechnet die Neighborhood Chunks
    //foreach point: chunk-ind-> list of pointindices
    let mut chunks: HashMap<(usize, usize), Vec<usize>> = HashMap::new();
    for j in 0..points.len() {
        let (indx, indy) = (
            (points[j].position.x / radius) as usize,
            (points[j].position.y / radius) as usize,
        );
        let entry = chunks.entry((indx, indy)).or_insert(Vec::new());
        entry.push(j);
        chunk_map.insert(j,Vec::new());
    }
    println!("Prestep NeighborhoodChunks complete");

    //Berechnet welche Chunks vollstaendig in Nbhd enthalten sind
    for x in 0..chunknmb {
        for y in 0..chunknmb {
            chunk_count.insert((x, y), 0);
            let lx = x as f64 / chunknmb as f64;
            let ux = (x + 1) as f64 / chunknmb as f64;
            let ly = y as f64 / chunknmb as f64;
            let uy = (y + 1) as f64 / chunknmb as f64;
            let (keyx, keyy) = ((lx / radius) as usize, (ly / radius) as usize);
            for j in 0..3  {
                for k in 0..3 {
                    if keyx + k > 0
                        && keyy + j > 0
                        && chunks
                            .contains_key(&(keyx + k - 1, keyy + j - 1))
                    {
                        chunks[&(keyx + k - 1, keyy + j - 1)]
                            .iter()
                            .for_each(|&p| {
                                if nbhd_check(points[p].position.distance_to(Point2D::new(lx, ly)),radius)
                                    && nbhd_check(points[p].position.distance_to(Point2D::new(lx, uy)),radius)
                                    && nbhd_check(points[p].position.distance_to(Point2D::new(ux, ly)),radius)
                                    && nbhd_check(points[p].position.distance_to(Point2D::new(ux, uy)),radius)
                                {
                                    let entry = chunk_map.entry(p).or_insert(Vec::new());
                                    entry.push((x, y));
                                }
                            })
                    }
                }
            }
        }
    }
    println!("Prestep chunkcontainment complete");

    //Berechnet die Punkte in der Nbhd
    points
        .par_iter()
        .enumerate()
        .map(|(i, point)| {
            // make sure that the point is in chunk_map
            let (keyx, keyy) = (
                (point.position.x / radius) as usize,
                (point.position.y / radius) as usize,
            );
            let mut neighbors: Vec<usize> = Vec::<usize>::new();
            for j in 0..3 {
                for k in 0..3 {
                    let neighbors_temp: Vec<usize> = if keyx + k > 0
                        && keyy + j > 0
                        && chunks.contains_key(&(keyx + k - 1, keyy + j - 1))
                    {
                        chunks[&(keyx + k - 1, keyy + j - 1)]
                            .iter()
                            .filter_map(|&p| {
                                let cx = (points[p].position.x * chunknmb as f64) as usize;
                                let cy = (points[p].position.y * chunknmb as f64) as usize;
                                if !chunk_map[&i].contains(&(cx, cy))
                                    && nbhd_check(points[p].position.distance_to(point.position), radius)
                                {
                                    Some(p)
                                } else {
                                    None
                                }
                            })
                            .collect()
                    } else {
                        Vec::<usize>::new()
                    };
                    neighbors.extend(neighbors_temp);
                }
            }
            (i, neighbors)
        })
        .collect::<Vec<_>>()
        .into_iter()
        .for_each(|(i, neighbors)| {
            neighbors_map.insert(i, neighbors);
        });
        println!("Prestep pointcontainment complete");

    for j in 0..points.len() {
        let cx = (points[j].position.x * chunknmb as f64) as usize;
        let cy = (points[j].position.y * chunknmb as f64) as usize;
        chunk_count.insert((cx, cy), chunk_count[&(cx, cy)] + 1);
    }
    //Calculates how many points are in each chunk
    (chunk_map, neighbors_map, chunk_count)
}

pub fn update_values(
    points: &mut Vec<DataPoint>,
    chunk_map: &HashMap<usize, Vec<(usize, usize)>>,
    neighbors: &HashMap<usize, Vec<usize>>,
    chunk_count: &HashMap<(usize, usize), i32>,
    chunknmb: usize,
) {
    //Chunk Sums berechnen
    let mut chunkvals: HashMap<(usize, usize), i32> = HashMap::new();

    let mut new_values:Vec<i32> = Vec::with_capacity(points.len());

    for x in 0..chunknmb {
        for y in 0..chunknmb {
            chunkvals.insert((x, y), 0);
        }
    }
    for j in 0..points.len() {
        if points[j].value == 1 {
            let (cx, cy) = (
                (points[j].position.x * chunknmb as f64) as usize,
                (points[j].position.y * chunknmb as f64) as usize,
            );
            chunkvals.insert((cx, cy), chunkvals[&(cx, cy)] + 1);
        }
    }

    new_values.par_extend((0..points.len()).into_par_iter().map(|i| {
        let neighbor_values: Vec<i32> = neighbors[&i]
            .iter()
            .map(|&j| points[j].value*2-1)
            .collect();
        let chunk_value_sum: Vec<i32> = chunk_map[&i]
            .iter()
            .map(|&(k, j)| chunkvals[&(k, j)] * 2 - chunk_count[&(k, j)])
            .collect();
        (chunk_value_sum.iter().sum::<i32>() + neighbor_values.iter().sum::<i32>()
            > 0_i32) as i32
    }));

    for (point, new_value) in points.iter_mut().zip(new_values.iter()) {
        point.value = *new_value;
    }
}

pub fn render_to_image(points: &[DataPoint], filename: &str, ellipse: &[(f64, f64)], conv_hull : &Vec<DataPoint>, ref_points :&Vec<usize>) {
    let width = 800;
    let height = 800;
    //let mut img = RgbImage::new(width, height);
    let mut img=RgbImage::from_pixel(width, height, Rgb([255, 255, 255]));

    for point in points {
        let color = if point.value == 1 {
            Rgb([59, 82, 139])
        } else {
            Rgb([253, 231, 37]) 
        };

        let x = (point.position.x * width as f64) as u32;
        let y = (point.position.y * height as f64) as u32;

        img.put_pixel(x, y, color);
    }

    let color = Rgb([255, 255, 255]);
    for i in ref_points {
        let x = (points[*i].position.x * width as f64) as u32;
        let y = (points[*i].position.y * height as f64) as u32;

        img.put_pixel(x, y, color);
    }

    /*let color = Rgb([255, 0, 0]);
                                  // the last point is the first point to close the loop
    let mut x0 = (ellipse[ellipse.len() - 1].0 * width as f64) as u32;
    let mut y0 = (ellipse[ellipse.len() - 1].1 * height as f64) as u32;

    // draw lines between points
    for point in ellipse {
        let x = (point.0 * width as f64) as u32;
        let y = (point.1 * height as f64) as u32;
        draw_line_segment_mut(
            &mut img,
            (x0 as f32, y0 as f32),
            (x as f32, y as f32),
            color,
        );
        x0 = x;
        y0 = y;
    }
    if conv_hull.len()>0 {
        let mut x0 = (conv_hull[conv_hull.len() - 1].position.x * width as f64) as u32;
        let mut y0 = (conv_hull[conv_hull.len() - 1].position.y * height as f64) as u32;

        let color = Rgb([0, 255, 0]); 
        // draw lines between points
        for point in conv_hull {
            let x = (point.position.x * width as f64) as u32;
            let y = (point.position.y * height as f64) as u32;
            draw_line_segment_mut(
                &mut img,
                (x0 as f32, y0 as f32),
                (x as f32, y as f32),
                color,
            );
            x0 = x;
            y0 = y;
        }
    }*/

    // let newrad: i32 = (radius * width as f64) as i32;
    // imageproc::drawing::draw_hollow_circle_mut(
    //     &mut img,
    //     (newrad * 2, newrad * 2),
    //     newrad,
    //     Rgb([255, 255, 255]),
    // );

    img.save(filename).unwrap();
}