#![allow(dead_code)]
use rayon::prelude::*;
use std::collections::HashMap;
use ndarray::prelude::*;
use image::{Rgb, RgbImage};
use imageproc::drawing::draw_line_segment_mut;
use crate::point::*;
use crate::nbhd_check;
use medians::*;//{*,algos::*}

use std::f64::consts::PI;

//Structure of File:
//compute_median -> Computes the median of given values
//  values: &[f64] -> f64

//UNUSED
//compute_mean -> Computes the mean of given values
//  values: &[f64] -> f64

//compute_neighborsf64 -> Computes nbhds
//  points: &[DataPointf64], radius: f64 -> HashMap<usize, Vec<usize>>

//fn update_valuesf64 -> Updates the values in each step
//  points: &mut Vec<DataPointf64>, neighbors: &HashMap<usize, Vec<usize>>, p_values : &Vec<f64>

//render_to_imagef64 -> Renders the image
//  points: &[DataPointf64], filename: &str, ellipse: &[(f64, f64)], ref_points :&Vec<usize>, _t:f64


//compute_p_valuesf64 -> Computes the area percentage for k-median
//points: &[DataPointf64], r : f64, angle:f64 -> &[f64]

pub fn compute_median(values: &[f64]) -> f64 {
    let mut sorted_values = values.to_vec();
    sorted_values.sort_by(|a, b| a.partial_cmp(b).unwrap());
    let mid: usize = sorted_values.len() / 2;

    if sorted_values.len() % 2 == 0 {
        (sorted_values[mid - 1] + sorted_values[mid]) / 2.0
    } else {
        sorted_values[mid]
    }
}

/*pub fn compute_p_median(values: &[f64], p_values: &Vec<f64>) -> f64 {
    let mut sorted_values = values.to_vec();
    sorted_values.sort_by(|a, b| a.partial_cmp(b).unwrap());
    let mid: usize = sorted_values.len() / 2;

    if sorted_values.len() % 2 == 0 {
        (sorted_values[mid - 1] + sorted_values[mid]) / 2.0
    } else {
        sorted_values[mid]
    }
}*/

pub fn compute_mean(values: &[f64]) -> f64 {
    values.iter().sum::<f64>()/values.len() as f64
}

pub fn compute_neighborsf64(
    points: &[DataPointf64],
    radius: f64,
) -> HashMap<usize, Vec<usize>> {
    //chunks and nbhdpunkte die bleiben
    let mut neighbors_map: HashMap<usize, Vec<usize>> = HashMap::new();

    //Berechnet die Neighborhood Chunks
    //foreach point: chunk-ind-> list of pointindices
    let mut chunks: HashMap<(usize, usize), Vec<usize>> = HashMap::new();
    for j in 0..points.len() {
        let (indx, indy) = (
            (points[j].position.x / radius) as usize,
            (points[j].position.y / radius) as usize,
        );
        let entry = chunks.entry((indx, indy)).or_insert(Vec::new());
        entry.push(j);
    }
    println!("Prestep NeighborhoodChunks complete");

    //Berechnet die Punkte in der Nbhd
    points
        .par_iter()
        .enumerate()
        .map(|(i, point)| {
            // make sure that the point is in chunk_map
            let (keyx, keyy) = (
                (point.position.x / radius) as usize,
                (point.position.y / radius) as usize,
            );
            let mut neighbors: Vec<usize> = Vec::<usize>::new();
            for j in 0..3 {
                for k in 0..3 {
                    let neighbors_temp: Vec<usize> = if keyx + k > 0
                        && keyy + j > 0
                        && chunks.contains_key(&(keyx + k - 1, keyy + j - 1))
                    {
                        chunks[&(keyx + k - 1, keyy + j - 1)]
                            .iter()
                            .filter_map(|&p| {
                                if nbhd_check(points[p].position.distance_to(point.position), radius)
                                {
                                    Some(p)
                                } else {
                                    None
                                }
                            })
                            .collect()
                    } else {
                        Vec::<usize>::new()
                    };
                    neighbors.extend(neighbors_temp);
                }
            }
            (i, neighbors)
        })
        .collect::<Vec<_>>()
        .into_iter()
        .for_each(|(i, neighbors)| {
            neighbors_map.insert(i, neighbors);
        });
        println!("Prestep pointcontainment complete");

    //Calculates how many points are in each chunk
    neighbors_map
}

pub fn update_valuesf64(
    points: &mut Vec<DataPointf64>,
    neighbors: &HashMap<usize, Vec<usize>>,
    p_values : &Vec<f64>,
) {
    let mut new_values:Vec<f64> = Vec::with_capacity(points.len());

    new_values.par_extend((0..points.len()).into_par_iter().map(|i| {
        let mut neighbor_values: Vec<f64> = neighbors[&i]
            .iter()
            .map(|&j| points[j].value as f64)
            .collect();
        
        //compute_median(&neighbor_values)
        
        //let a=p_values[i];
        let val:f64;
        if p_values[i]>1.
        {
            val=-100.;
        }else{
            if p_values[i]< -1.
            {
                val=100.;
            }else{
                let len=neighbor_values.len() as f64;
                let (_lesser, median, _greater) 
                    = neighbor_values.select_nth_unstable_by(((1f64+p_values[i])/2f64*(len-1.))as usize, |a, b| b.total_cmp(a));
                val=*median;
            }
        }
        val
        //compute_p_median(&neighbor_values, &p_values)
        
        //(&neighbor_values).medf_checked().unwrap()

        //compute_mean(&neighbor_values)
        //neighbor_values.medf_unchecked()
    }));

    for (point, new_value) in points.iter_mut().zip(new_values.iter()) {
        point.value = *new_value;
    }
}

pub fn render_to_imagef64(points: &[DataPointf64], filename: &str, ref_points :&Vec<usize>, _t:f64, ellipse: &[(f64, f64)]) {
    let width = 800;
    let height = 800;
    //let mut img = RgbImage::new(width, height);
    let mut img=RgbImage::from_pixel(width, height, Rgb([255, 255, 255]));

    for point in points {
        // let color = if point.value >0.16 {
        //     Rgb([59, 82, 139])
        // } else {
        //     Rgb([253, 231, 37]) 
        // };

        let color;
        color=match point.value {
            x if x > 50. => Rgb([255,0,0]),
            x if x > 0.4 => Rgb([61,38,168]),
            x if x > 0.3 => Rgb([71,87,247]),
            x if x > 0.25 => Rgb([39,150,235]),
            x if x > 0.2 => Rgb([24,191,181]),
            x if x > 0.16 => Rgb([128,203,88]),
            x if x > 0.1 => Rgb([253,189,60]),
            x if x > -50. => Rgb([249,250,20]),
            _ => Rgb([0,0,255]),
        };

        // let mut scal=100.*(point.value-point_valuef64(point.position.x,point.position.y)-4.*t);
        // scal=scal.abs().min(1.);
        // let color=Rgb([(253.*scal+59.*(1.-scal)) as u8, (231.*scal+82.*(1.-scal)) as u8, (37.*scal+139.*(1.-scal)) as u8]);

        let x = (point.position.x * width as f64) as u32;
        let y = (point.position.y * height as f64) as u32;

        img.put_pixel(x, y, color);
    }

    let color = Rgb([255, 255, 255]);
    for i in ref_points {
        let x = (points[*i].position.x * width as f64) as u32;
        let y = (points[*i].position.y * height as f64) as u32;

        img.put_pixel(x, y, color);
    }

    let color = Rgb([255, 0, 0]);//Rgb([255, 0, 0])
    //the last point is the first point to close the loop
    let mut x0 = (ellipse[ellipse.len() - 1].0 * width as f64) as u32;
    let mut y0 = (ellipse[ellipse.len() - 1].1 * height as f64) as u32;

    //draw lines between points
    for point in ellipse {
        let x = (point.0 * width as f64) as u32;
        let y = (point.1 * height as f64) as u32;
        draw_line_segment_mut(
            &mut img,
            (x0 as f32, y0 as f32),
            (x as f32, y as f32),
            color,
        );
        x0 = x;
        y0 = y;
    }

    img.save(filename).unwrap();
}

pub fn compute_p_valuesf64(
    points: &[DataPointf64], 
    r : f64, 
    angle:f64,
) -> Vec<f64> {//HashMap<usize, f64>
let mut p_values=Vec::<f64>::new();

points
        .par_iter()
        .enumerate()
        .map(|(i, point)| {

            let dx=(point.position.x/r).min((1.-point.position.x)/r);
            let dy=(point.position.y/r).min((1.-point.position.y)/r);
            let mut ax=0f64; 
            let mut ay=0f64;
            let mut volx=0f64;
            let mut voly=0f64;
            if dx<1f64 {
                ax=dx.acos(); 
                volx=ax-(1f64-dx*dx).sqrt()*dx;
            }
            if dy<1f64 {
                ay=dy.acos(); 
                voly=ay-(1f64-dy*dy).sqrt()*dy;
            }
            let mut p_vol=volx+voly;
            if dx<1f64 && dy<1f64 && ax+ay>PI/2f64 {
                p_vol=p_vol-(volx+voly-PI/2f64)/2f64-dx*dy;
            }
            //vorher angle*p_vol/PI
            p_vol=(angle*(p_vol)/(PI-p_vol)).min(1.).max(-1.);
            //p_vol=angle*(p_vol)/(PI-p_vol);

            (i, p_vol)
        })
        .collect::<Vec<_>>()
        .into_iter()
        .for_each(|(i, p_value)| {
            p_values.insert(i, p_value);
        });

p_values
}

