use euclid::Point2D;
use rand::Rng;
pub enum EmptyUnit {}

//Structure of File:
//DataPoint -> The Point structure to represent data
//  position: Point2D<f64,EmptyUnit>, value: i32
// point_value x: f64, y: f64 -> i32
// generate_random_points n: usize -> Vec<DataPoint>

//DataPointf64 -> The Point structure to represent data
//  position: Point2D<f64,EmptyUnit>, value: f64
// point_value x: f64, y: f64 -> f64
// generate_random_points n: usize -> Vec<DataPointf64>

#[derive(Debug, Clone, Copy)]
pub struct DataPoint {
    pub position: Point2D<f64,EmptyUnit>,
    pub value: i32,
}

pub fn point_value(x: f64, y: f64) -> i32 {
    //Ellipse
    //((x - 0.5).powi(2) / 4.0 + (y - 0.5).powi(2) > 0.02) as i32

    //Circle
    //((x - 0.5).powi(2) + (y - 0.5).powi(2) > 0.16) as i32

    //Random
    //(x<0.5) as i32
    //(i%2==0) as i32
    //(1.*(x-0.5)+10.*(y-0.5)>0.) as i32
    (1.*(x-0.5)+(y-0.5)>0.) as i32
    //rand::random::<bool>() as i32
}

// Generate `n` random points in the unit square
pub fn generate_random_points(n: usize) -> Vec<DataPoint> {
    let mut rng = rand::thread_rng();
    (0..n)
        .map(|i| {
            //random
            //let x : f64 = rng.gen_range(0.0..1.0);
            //let y : f64 = rng.gen_range(0.0..1.0);
            //two spheres
            let mut x : f64;
            let mut y : f64;
            let choice=rng.gen_range(0.0..1.0);
            let angle=rng.gen_range(0.0..2.0*std::f64::consts::PI);
            let angle2=rng.gen_range(0.0..2.0*std::f64::consts::PI);
            let dist=rng.gen_range(0.0..1.0);
            let pert=rng.gen_range(0.0..1.0);
            if choice <0.1
            {
                x=0.45+dist*0.1;//45 bis 55
                y=0.5;
            }else{
                if choice<0.55
                {
                    x=0.15*angle.cos()+0.3;
                    y=0.15*angle.sin()+0.5;
                }else{
                    x=0.15*angle.cos()+0.7;
                    y=0.15*angle.sin()+0.5;
                }
            }
            x=x+angle2.cos()*pert*0.05;
            y=y+angle2.sin()*pert*0.05;
            let value = point_value(x, y);
            DataPoint { position: Point2D::new(x, y), value }
        })
        .collect()
}

#[derive(Debug, Clone, Copy)]
pub struct DataPointf64 {
    pub position: Point2D<f64,EmptyUnit>,
    pub value: f64,
}

pub fn point_valuef64(x: f64, y: f64) -> f64 {
    //(x - 0.5).powi(2) + (y - 0.5).powi(2)
    (6.*(x - 0.5).powi(2) / 4.0 + 6.*(y - 0.5).powi(2)).min(0.3)

}

// Generate `n` random points in the unit square
pub fn generate_random_pointsf64(n: usize) -> Vec<DataPointf64> {
    let mut rng = rand::thread_rng();
    (0..n)
        .map(|_| {
            let x : f64 = rng.gen_range(0.0..1.0);
            let y : f64 = rng.gen_range(0.0..1.0);
            let value = point_valuef64(x, y);
            DataPointf64 { position: Point2D::new(x, y), value }
        })
        .collect()
}